#!/usr/bin/env python

# Copyright (C) 2002 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import threading
import Queue
import ao

import events
import pcm
import mp3file
from player import genericplayer

from helper import debug

class bufferedao(threading.Thread):
    def __init__(self, aodevice, rate, aooptions, queuesize):
        self.ao = ao.AudioDevice(aodevice, rate=rate, options=aooptions)
        self.queue = Queue.Queue(queuesize)
        self.done = 0
        threading.Thread.__init__(self)
        
        # we don't know what to do, if we are left alone, hence:
        self.setDaemon(1)
        
    def run(self):
        while not self.done:
            buff, bytes = self.queue.get(1)
            if buff!=0 and bytes!=0:
                self.ao.play(buff, bytes)

    def play(self, buff, bytes):
        self.queue.put((buff, bytes))

    def stop(self):
        self.done = 1
        while self.queue.full():
            self.queue.get(1)
        self.play(0,0)


class player(genericplayer):

    def __init__(self,
                 aodevice="oss", aooptions={},
                 bufsize=0,
                 autoplay=0,
                 crossfadingstart=5,
                 crossfadingduration=10):
        
        self.rate = 44100
        self.SIZE = 4806
        if bufsize>0:
            queuesize = 1024*bufsize/self.SIZE + 1
            self.ao = bufferedao(aodevice,
                                 self.rate, 
                                 aooptions,
                                 queuesize)

            self.ao.start()
            
            # calulate maximal length of pcm samples in queue in 1/1000th seconds
            self.queuelength = 1000/4.0*queuesize*self.SIZE/self.rate
        else:
            self.ao = ao.AudioDevice(aodevice, rate=self.rate, options=aooptions)
            self.queuelength = 0

        # songs currently playing
        self.songs = []

        # crossfading parameters (all values in seconds and change/second, resp.)
        self.crossfadingduration = crossfadingduration
        self.crossfadingstart = crossfadingstart
        self.crossfadingratio = 0
        self.crossfadingrate = 1.0/self.rate/self.crossfadingduration

        genericplayer.__init__(self, autoplay)

    def play(self):
        """decode songs and mix them together"""
        if len(self.songs)==1:
            song = self.songs[0]
            buff = song.read(self.SIZE)
            if len(buff)>0:
                self.ao.play(buff, len(buff))
            else:
                debug("madplayer: song ends: %s (0 songs in queue)\n" % self.songs[0])
                del self.songs[0]

            if (song.ttime-song.ptime)/1000<self.crossfadingstart: 
                self.requestnextsong()

        elif len(self.songs)==2:
            if self.crossfadingratio is not None:
                # perform crossfading
                buff1 = self.songs[0].read(self.SIZE)
                buff2 = self.songs[1].read(self.SIZE)

                if len(buff1) and len(buff2):
                    # normal operation: no song has ended
                    buff, self.crossfadingratio = pcm.mix(buff1, buff2,
                                                          self.crossfadingratio,
                                                          self.crossfadingrate)
                    if self.crossfadingratio>=1:
                        self.crossfadingratio = 0
                        debug("madplayer: song ends: %s (1 song in queue)\n" %
                              self.songs[0])
                        del self.songs[0]
                if len(buff1)==0:
                    buff = buff2
                    self.crossfadingratio = 0
                    debug("madplayer: song ends: %s (1 song in queue)\n" % self.songs[0])
                    del self.songs[0]
                if len(buff2)==0:
                    buff = buff1
                    self.crossfadingratio = 0
                    debug("madplayer: song ends: %s" % self.songs[-1])
                    del self.songs[-1]
                    debug(" (%d songs in queue)\n" % len(self.songs))
            else:
                # just kill gap between songs
                buff = self.songs[0].read(self.SIZE)
                if len(buff)<self.SIZE:
                    self.crossfadingratio = 0
                    debug("madplayer: song ends: %s (1 song in queue)\n" % self.songs[0])
                    del self.songs[0]
                    
                    buff2 = self.songs[0].read(self.SIZE)
                    if len(buff2)==0:
                        debug("madplayer: song ends: %s" % self.songs[0])
                        del self.songs[0]
                        debug(" (%d songs in queue)\n" % len(self.songs))
                    else:
                        buff = buff + buff2

            if len(buff)>0:
                self.ao.play(buff, len(buff))

        # update playbackinfo

        if len(self.songs)>0:
            # determine which song is currently played
            # try to take the buffer length into account
            if self.crossfadingratio and self.songs[-1].ptime-self.queuelength>=0:
                playingsong = self.songs[-1]
                title = "-> %s" % playingsong.title
            else:
                playingsong = self.songs[0]
                title = "%s" % playingsong.title
                
            ptime = int(max(playingsong.ptime-self.queuelength, 0)/1000)
            ttime = int(playingsong.ttime/1000)
            self.playbackinfo = (title, ptime, ttime)
        else:
            self.playbackinfo = ("", None, None)

    def _nextsong(self, song):
        debug("madplayer: new song: %s\n" % song)

        # we want maximally 2 songs in queue
        if len(self.songs)==2:
            del self.songs[0]

        try:
            self.songs.append(mp3file.mp3file(song, self.rate))

            # Check whether two songs come after each other on an
            # album In such a case, we don't want to crossfade.  If,
            # however, the user has requested the song change (via
            # playerforward), we do want do crossfade.
            # Whether this has been the case, can we infered from
            # self.autorequestednextsong.
            if self.autorequestednextsong and len(self.songs)==2:
                song1 = self.songs[0].song
                song2 = self.songs[1].song
                # XXX no check for artist/album==UNKNOWN!
                if (song1.artist==song2.artist and
                    song1.album==song2.album and
                    song1.tracknr!="" and song2.tracknr!="" and
                    int(song1.tracknr)==int(song2.tracknr)-1):
                    self.crossfadingratio = None
                    debug("madplayer: I don't crossfade successive songs.\n")
                
        except IOError: pass

        debug(" (%d songs in queue)\n" % len(self.songs))

    def _playerstop(self):
        self.songs = []


